from __future__ import annotations
import importlib


def index_to_type(index: int):
    match index:
        case 3:
            # lazy import to prevent dep cycle
            mod = importlib.import_module("irys_sdk.bundle.signers.ethereum")
            return mod.EthereumSigner
            # from irys_sdk.bundle.signers.ethereum import EthereumSigner
            # return EthereumSigner



from time import sleep
from urllib.parse import urljoin
import requests
from irys_sdk.utils import check_and_throw, confirmation_poll, get_bundler_address

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from irys_sdk.client import Uploader


class Fund:
    irys: "Uploader"

    def __init__(self, irys):
        self.irys = irys

    def submit_transaction(self, tx_id: str):
        submit_url = urljoin(
            self.irys.url, "/account/balance/{}".format(self.irys.token))
        for _i in range(0, 4, 1):
            try:
                res = requests.post(submit_url, json={'tx_id': tx_id})
                check_and_throw(res, "submitting fund tx")
                return res
            except:
                sleep(2)
                pass
        raise Exception("Node was unable to confirm tx {}".format(tx_id))

    def fund(self, amount: int, multiplier=1.0):
        to = get_bundler_address(self.irys.url, self.irys.token_name)
        tx = self.irys.token.create_tx(amount, to)
        send_res = self.irys.token.send_tx(tx)
        tx_id = tx["tx_id"] if "tx" in tx else send_res  # type: ignore
        confirmation_poll(self.irys, tx_id)
        self.submit_transaction(tx_id)
        return send_res



def byte_array_to_long(byte_array: bytearray):
    value = 0
    for i in range(len(byte_array)-1, -1, -1):
        value = value * 256 + byte_array[i]
    return value


def long_to_n_byte_array(N: int, long: int) -> bytearray:
    byte_array = bytearray(N)
    if (long < 0):
        raise Exception("Array is unsigned, cannot represent -ve numbers")
    if (long > 2 ** (N * 8) - 1):
        raise Exception(
            "Number {} is too long to represent in an array of {} bytes".format(long, N))
    for index in range(0, len(byte_array), 1):
        byte = long & 0xff
        byte_array[index] = byte
        long = (long - byte) // 256

    return byte_array


def short_to_2_byte_array(long: int) -> bytearray:
    return long_to_n_byte_array(2, long)


def long_to_8_byte_array(long: int) -> bytearray:
    return long_to_n_byte_array(8, long)


def long_to_16_byte_array(long: int) -> bytearray:
    return long_to_n_byte_array(16, long)


def long_to_32_byte_array(long: int) -> bytearray:
    return long_to_n_byte_array(32, long)


def set_bytes(dest: bytearray, src: bytearray, offset: int):
    for i in range(offset, offset + len(src), 1):
        dest[i] = src[i - offset]



from __future__ import annotations
from typing import Literal
from irys_sdk.client import SupportedTokens, Uploader


class Builder:
    _node_url = "https://uploader.irys.xyz"
    _network = "mainnet"
    token_opts = {"provider_url": None}

    def __init__(self, token: SupportedTokens):
        self.token = token

    @property
    def network(self):
        self._network

    def network(self, network: Literal["mainnet", "devnet"]):
        match network:
            case "mainnet":
                self._node_url = "https://uploader.irys.xyz"
            case "devnet":
                self._node_url = "https://devnet.irys.xyz"
            case _:
                raise Exception(
                    "invalid network, expected 'mainnet' or 'devnet'")

        self._network = network
        return self

    # @property
    # def rpc_url(self):
    #     self.token_opts["provider_url"]

    def rpc_url(self, rpc_url: str):
        self.token_opts["provider_url"] = rpc_url
        return self

    # @property
    # def wallet(self):
    #     return self.token_opts["wallet"]

    def wallet(self, wallet: str):
        self.token_opts["wallet"] = wallet
        return self

    def node_url(self, node_url: str):
        self._node_url = node_url
        return self

    # @property
    # def url(self) -> str:
    #     return self._node_url

    def build(self) -> "Uploader":
        if (self.network == "devnet" and self.rpc_url == None):
            raise Exception("rpc url must be provided when using devnet")

        uploader = Uploader(self._node_url, self.token,
                            token_opts=self.token_opts)
        return uploader



from time import sleep
from urllib.parse import urljoin
from warnings import warn

import requests


def get_balance(url: str, token: str, address: str):
    balance_url = urljoin(
        url, "/account/balance/{}?address={}".format(token, address))
    res = requests.get(balance_url)
    check_and_throw(res, "getting balance")
    return int(res.json()["balance"])


def get_price(url: str, token: str, bytes: int):
    balance_url = urljoin(
        url, "/price/{}/{}".format(token, str(bytes)))
    res = requests.get(balance_url)
    check_and_throw(res, "getting price")
    return int(res.text)


def get_bundler_address(url: str, token: str):
    res = requests.get(url)
    check_and_throw(res, "getting price")
    return res.json()["addresses"][token]


def confirmation_poll(irys, tx_id):
    # TODO @JesseTheRobot - impl proper tx conf checking logic
    for _i in range(1, 60, 1):
        try:
            return irys.token.get_tx(tx_id)
        except:
            pass
        sleep(0.5)
    warn("unable to confirm tx {} in 30s".format(tx_id))
    return


def check_and_throw(response, context="unknown", exceptions=[]):
    if response.status_code in exceptions:
        return
    if response.status_code != 200:
        raise Exception("HTTP Error while {} : {} - {}".format(context,
                                                               response.status_code, response.content))



from __future__ import annotations
from typing import Any
from eth_account import Account
from web3 import Web3
from irys_sdk.bundle.signers.ethereum import EthereumSigner
from irys_sdk.tokens.base import BaseToken
from eth_hash.auto import keccak

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from irys_sdk.client import Uploader


class EthereumToken(BaseToken):
    irys: "Uploader" = None
    address: str = None
    _provider_instance = None
    name = "ethereum",
    ticker = "ETH"
    base = ("wei", int(1e18))

    def __init__(self, irys, **kwargs):
        self.irys = irys

        self.provider_url = kwargs.get(
            "provider_url") or "https://cloudflare-eth.com/"
        self.name = kwargs.get("name") or "ethereum"
        self.ticker = kwargs.get("ticker") or "ETH"

        self._wallet = kwargs.get("wallet")
        if self._wallet == None:
            raise Exception("missing required wallet arg")
        pass

    def get_provider(self):
        if self._provider_instance == None:
            self._provider_instance = Web3(
                Web3.HTTPProvider(self.provider_url))
        return self._provider_instance

    def get_public_key(self) -> bytearray:
        return self.get_signer().public_key

    def get_signer(self):
        return EthereumSigner(self._wallet)

    def get_tx(self, tx_id: str):
        provider = self.get_provider()
        return provider.eth.get_transaction(tx_id)

    def owner_to_address(self, pub: bytearray) -> str:
        pubkey = pub if len(pub) == 64 else pub[1:]
        kek = keccak(pubkey)
        return "0x" + kek[-20:].hex()

    def ready(self):
        self.address = self.owner_to_address(self.get_public_key())

    def create_tx(self, amount: int, to: str, fee=None):
        provider = self.get_provider()
        chain_id = provider.eth.chain_id
        nonce = provider.eth.get_transaction_count(
            Web3.to_checksum_address(self.address))
        txb = {
            'to': to,
            'value': amount,
            "gas": 100000,
            "maxFeePerGas": 2000000000,
            "maxPriorityFeePerGas": 2000000000,
            'nonce': nonce,
            'chainId': chain_id
        }

        signed_tx = Account.sign_transaction(txb, self._wallet)

        return (signed_tx)

    def send_tx(self, tx: Any) -> Any:
        provider = self.get_provider()
        # res = provider.eth.send_raw_transaction(tx.raw_transaction)
        # return res.to_0x_hex()

        res = provider.eth.send_raw_transaction(tx.rawTransaction)
        return res.hex()
